﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace mjs {
    /* GUI関係の処理はこちら */
    public partial class MainForm : Form {
        /* 牌の表示関係 */
        public const int HAI_WIDTH = 24, HAI_HEIGHT = 34;
        public static Rectangle[] HaiToRect;
        //得点表示ウインドウ
        FormScore score_win;

        /* 現在の状態 */
        private MJState state = new MJState();

        /* 現在の牌選択用フォーカス先 */
        enum FocusCtl {
            EMPTY,
            OMOTE_DORA,
            URA_DORA,
            JIHAI,
            NAKI1,
            NAKI2,
            NAKI3,
            NAKI4,
            AGARI_HAI,
        };
        private FocusCtl focus = FocusCtl.EMPTY;

        /* 牌をソートする述語関数 */
        private static int HaiCompare(Hai l, Hai r) {
            return Math.Sign(l - r);
        }



        /* 牌の描画 */
        delegate void DrawAct(int x, int y, Hai h);
        private void Redraw() {
            int i, j;

            //再描画

            Bitmap bmp = new Bitmap(HAI_WIDTH * 13, HAI_HEIGHT);
            Image data = picHaiList.Image;
            Graphics g = Graphics.FromImage(bmp);

            //1個の牌を描く
            DrawAct DrawHai = delegate(int x, int y, Hai h) {
                g.DrawImage(data, new Rectangle(x, y, HAI_WIDTH, HAI_HEIGHT), HaiToRect[(int)h], GraphicsUnit.Pixel);
            };


            //表ドラ
            g.Clear(Color.White);
            for (i = 0; i < 5; i++) {
                if (state.OmoteDora.Count <= i) break;
                DrawHai(HAI_WIDTH * i, 0, state.OmoteDora[i]);
            }
            if (focus == FocusCtl.OMOTE_DORA)
                g.FillRectangle(new SolidBrush(Color.FromArgb(64, Color.Purple)), 0, 0, HAI_WIDTH * 5, HAI_HEIGHT);
            picOmoteDora.CreateGraphics().DrawImage(bmp, 0, 0);

            //裏ドラ
            g.Clear(Color.White);
            for (i = 0; i < 5; i++) {
                if (state.UraDora.Count <= i) break;
                DrawHai(HAI_WIDTH * i, 0, state.UraDora[i]);
            }
            if (focus == FocusCtl.URA_DORA)
                g.FillRectangle(new SolidBrush(Color.FromArgb(64, Color.Purple)), 0, 0, HAI_WIDTH * 5, HAI_HEIGHT);
            picUraDora.CreateGraphics().DrawImage(bmp, 0, 0);

            //手持ち牌
            g.Clear(Color.White);
            j = 13 - state.ArrayNaki.Count * 3;
            for (i = 0; i < j; i++) {
                if (state.Jihai.Count <= i) break;
                DrawHai(HAI_WIDTH * i, 0, state.Jihai[i]);
            }
            if (focus == FocusCtl.JIHAI)
                g.FillRectangle(new SolidBrush(Color.FromArgb(64, Color.Purple)), 0, 0, HAI_WIDTH * (13 - state.ArrayNaki.Count * 3), HAI_HEIGHT);

            for (i = 0; i < state.ArrayNaki.Count; i++) {
                int pos = j + 3 * i;
                if (state.ArrayNaki[i].top != Hai.EMPTY) {
                    Mentsu_State ns = state.ArrayNaki[i];

                    switch (ns.type) {
                    case Mentsu.SHUNZU: //チー
                        DrawHai(HAI_WIDTH * pos, 0, ns.top);
                        DrawHai(HAI_WIDTH * (pos + 1), 0, ns.top + 1);
                        DrawHai(HAI_WIDTH * (pos + 2), 0, ns.top + 2);
                        break;

                    case Mentsu.KOHZU: //ポン
                        DrawHai(HAI_WIDTH * pos, 0, ns.top);
                        DrawHai(HAI_WIDTH * (pos + 1), 0, ns.top);
                        DrawHai(HAI_WIDTH * (pos + 2), 0, ns.top);
                        break;

                    case Mentsu.KANZU: //明カン・暗カン
                        if(ns.naki) {
                            //明カン
                            DrawHai(HAI_WIDTH * pos, 0, ns.top);
                            DrawHai(HAI_WIDTH * (pos + 1), 0, ns.top);
                            DrawHai(HAI_WIDTH * (pos + 2), 0, ns.top);
                            DrawHai(HAI_WIDTH * (pos + 1), 4, ns.top);
                        }
                        else {
                            //暗カン
                            DrawHai(HAI_WIDTH * pos, 0, ns.top);
                            DrawHai(HAI_WIDTH * (pos + 1), 0, ns.top);
                            DrawHai(HAI_WIDTH * (pos + 2), 0, ns.top);
                            DrawHai(HAI_WIDTH * (pos + 1) , 4, Hai.URA);
                        }
                        break;
                    }


                }
                if ((focus - FocusCtl.NAKI1) == i) {
                    g.FillRectangle(new SolidBrush(Color.FromArgb(64, Color.Purple)), HAI_WIDTH * (j + 3 * i), 0, HAI_WIDTH * 3, HAI_HEIGHT);
                }
            }

            picJihai.CreateGraphics().DrawImage(bmp, 0, 0);

            //上がり牌
            g.Clear(Color.White);
            DrawHai(0, 0, state.AgariHai);
            if (focus == FocusCtl.AGARI_HAI)
                g.FillRectangle(new SolidBrush(Color.FromArgb(64, Color.Purple)), 0, 0, HAI_WIDTH, HAI_HEIGHT);
            picAgariHai.CreateGraphics().DrawImage(bmp, 0, 0);

        }


        private void FixNumNaki() {
            //鳴きの数の調整
            Label[] ctl = { lblNaki1, lblNaki2, lblNaki3, lblNaki4 };
            String[] str = { "└  チー  ┘", "└  ポン  ┘", "└ 明カン ┘", "└ 暗カン ┘" };
            List<Mentsu_State>[] oldState = new List<Mentsu_State>[4];
            int[] newNum = new int[4], newOffset = new int[4];
            int[] type = new int[4];

            //鳴き牌
            int max = 0, i, j;
            max = (int)(numChi.Value + numPon.Value + numAnkan.Value + numMinkan.Value);
            if (max > 4) max = 4;

            //元データの保存
            for (i = 0; i < 4; i++) {
                oldState[i] = new List<Mentsu_State>();
            }
            foreach (Mentsu_State n in state.ArrayNaki) {
                int t = 0;
                if (n.type == Mentsu.SHUNZU) t=0; //チー
                else if (n.type == Mentsu.KOHZU) t=1;
                else if (n.type == Mentsu.KANZU && n.naki) t = 2;
                else if (n.type == Mentsu.KANZU && !n.naki) t = 3;

                oldState[t].Add(n);
            }

            //新しい鳴きの数
            newNum[0] = (int)numChi.Value;
            newNum[1] = (int)numPon.Value;
            newNum[2] = (int)numMinkan.Value;
            newNum[3] = (int)numAnkan.Value;
            newOffset[0] = 0;
            for (j = 1; j < 4; j++) newOffset[j] = newOffset[j - 1] + newNum[j - 1];

            for (i = 0; i < max; i++) {
                for (j = 0; j < 4; j++) {
                    if (i >= newOffset[j] && i < newOffset[j] + newNum[j]) type[i] = j;
                }
            }

            //更新及びデータの充当
            state.ArrayNaki.Clear();
            for (i = 0; i < 4; i++) {
                ctl[i].Visible = (4-i <= max);
            }
            for (i = 0; i < max; i++) {

                ctl[4 - max + i].Text = str[type[i]];

                if (oldState[type[i]].Count > 0) {
                    state.ArrayNaki.Add(oldState[type[i]][0]);
                    oldState[type[i]].RemoveAt(0);
                }
                else {
                    Mentsu_State ns = new Mentsu_State();
                    if (type[i] == 0) { ns.type = Mentsu.SHUNZU; ns.naki = true; } //チー
                    if (type[i] == 1) { ns.type = Mentsu.KOHZU; ns.naki = true; } //ポン
                    if (type[i] == 2) { ns.type = Mentsu.KANZU; ns.naki = true; } //明カン
                    if (type[i] == 3) { ns.type = Mentsu.KANZU; ns.naki = false; } //暗カン

                    ns.top = Hai.EMPTY;
                    state.ArrayNaki.Add(ns);
                }
            }
            //鳴きが増えた場合、古いデータを削除
            i = 13 - max * 3;
            if (state.Jihai.Count > i)
                state.Jihai.RemoveRange(i, state.Jihai.Count - i);

            Redraw();
        }

        public MainForm() {
            InitializeComponent();
            int i;

            /* 牌画像位置の初期化 */
            HaiToRect = new Rectangle[41];
            for (i = 0; i < HaiToRect.Length; i++) HaiToRect[i] = new Rectangle(0,0,0,0);
            for (i = 1; i < 10; i++) {
                HaiToRect[0 + i] = new Rectangle(HAI_WIDTH * (i-1), 0 , HAI_WIDTH, HAI_HEIGHT);
                HaiToRect[10 + i] = new Rectangle(HAI_WIDTH * (i-1), HAI_HEIGHT, HAI_WIDTH, HAI_HEIGHT);
                HaiToRect[20 + i] = new Rectangle(HAI_WIDTH * (i-1), HAI_HEIGHT*2, HAI_WIDTH, HAI_HEIGHT);
            }
            HaiToRect[31] = new Rectangle(HAI_WIDTH * 0, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);
            HaiToRect[32] = new Rectangle(HAI_WIDTH * 1, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);
            HaiToRect[33] = new Rectangle(HAI_WIDTH * 2, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);
            HaiToRect[34] = new Rectangle(HAI_WIDTH * 3, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);
            HaiToRect[36] = new Rectangle(HAI_WIDTH * 4, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);
            HaiToRect[37] = new Rectangle(HAI_WIDTH * 5, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);
            HaiToRect[38] = new Rectangle(HAI_WIDTH * 6, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);
            HaiToRect[40] = new Rectangle(HAI_WIDTH * 7, HAI_HEIGHT * 3, HAI_WIDTH, HAI_HEIGHT);

        }
        
        private void MainForm_Load(object sender, EventArgs e) {
            lstJikaze.SelectedIndex = 0;
            lstBakaze.SelectedIndex = 0;
            lstRichi.SelectedIndex = 0;
            lstTsumo.SelectedIndex = 0;
            score_win = new FormScore();
        }

        private void numChi_ValueChanged(object sender, EventArgs e) {
            /* チー・ポン・カンの数が変わった場合 */
            FixNumNaki();
        }

        private void btnCalc_Click(object sender, EventArgs e) {
            //計算実行
            List<MJState_Result> res;
            MJState_Resolver mjs = new MJState_Resolver();
            state.Bakaze = (Kaze)lstBakaze.SelectedIndex;
            state.Jikaze = (Kaze)lstJikaze.SelectedIndex;
            state.TsumoAgari = (lstTsumo.SelectedIndex==0);
            state.Richi = lstRichi.SelectedIndex;
            state.Haitei = chkHaitei.Checked;
            state.Ippatsu = chkIppatsu.Checked;
            state.Rinshan = chkRinshan.Checked;
            state.Chankan = chkChankan.Checked;

            res = mjs.Resolve(state);
            if(mjs.ErrorStr != "") {
                lblResult.Text = mjs.ErrorStr;
            }
            else if(res.Count == 0) {
                lblResult.Text = "上がりパターンが見つかりませんでした。";
            }
            else {
                score_win.SetData(res,picHaiList.Image);
                score_win.ShowDialog();
                Redraw();
            }

        }

        private void picOmoteDora_MouseDown(object sender, MouseEventArgs e) {
            //表ドラクリック
            if (e.Button == MouseButtons.Right) {
                int x = e.X / HAI_WIDTH;
                if (x < state.OmoteDora.Count) state.OmoteDora.RemoveAt(x);
            }
            focus = FocusCtl.OMOTE_DORA;
            Redraw();

        }

        private void picUraDora_MouseDown(object sender, MouseEventArgs e) {
            //裏ドラクリック
            if (e.Button == MouseButtons.Right) {
                int x = e.X / HAI_WIDTH;
                if (x < state.UraDora.Count) state.UraDora.RemoveAt(x);
            }
            focus = FocusCtl.URA_DORA;
            Redraw();

        }

        private void picJihai_MouseDown(object sender, MouseEventArgs e) {
            int x = e.X / HAI_WIDTH;
            int i = 13 - state.ArrayNaki.Count * 3;
            //自牌クリック
            if (e.Button == MouseButtons.Right) {
                if (x < state.Jihai.Count) {
                    state.Jihai.RemoveAt(x);
                }
                if (x >= i && x < i + 3 * state.ArrayNaki.Count) {
                    Mentsu_State ns = state.ArrayNaki[(x - i) / 3];
                    ns.top = Hai.EMPTY;
                    state.ArrayNaki[(x - i) / 3] = ns;
                }
            }
            if (x < i) {
                focus = FocusCtl.JIHAI;
            }
            else if (x >= i && x < i + 3 * state.ArrayNaki.Count) {
                focus = FocusCtl.NAKI1 + ((x - i) / 3);
            }

            Redraw();

        }

        private void picAgariHai_MouseDown(object sender, MouseEventArgs e) {
            //上がり牌クリック
            if (e.Button == MouseButtons.Right) state.AgariHai = Hai.EMPTY;
            focus = FocusCtl.AGARI_HAI;
            Redraw();

        }

        private void picHaiList_MouseDown(object sender, MouseEventArgs e) {
            //牌のリストをクリック
            int i, j;

            Hai target;
            j = -1;
            for (i = 1; i < HaiToRect.Length; i++) {
                if (HaiToRect[i].Contains(e.Location) && i < (int)Hai.URA) {
                    j = i;
                    break;
                }
            }
            if (j == -1) return;
            if (focus == FocusCtl.EMPTY) return;
            target = (Hai)j;
            if (target >= Hai.URA) return;

            if (focus >= FocusCtl.NAKI1 && focus - FocusCtl.NAKI1 < state.ArrayNaki.Count) {

                Mentsu_State ns = state.ArrayNaki[focus - FocusCtl.NAKI1];
                ns.top = target;
                

                switch (ns.type) {
                case Mentsu.SHUNZU: //チー
                    //チーは1～7・11～17・21～27のみ選択可能
                    if ((j % 10) < 1 || (j % 10) > 7) return;
                    if ((j / 10) < 0 || (j / 10) > 2) return;
                    break;
                default:
                    //チー以外は何でも可
                    break;
                }
                state.ArrayNaki[focus - FocusCtl.NAKI1] = ns;

            }
            else {

                if (focus == FocusCtl.OMOTE_DORA) {
                    if (state.OmoteDora.Count > 4) return;
                    state.OmoteDora.Add(target);
                    state.OmoteDora.Sort(HaiCompare);
                }
                else if (focus == FocusCtl.URA_DORA) {
                    if (state.UraDora.Count > 4) return;
                    state.UraDora.Add(target);
                    state.UraDora.Sort(HaiCompare);
                }
                else if (focus == FocusCtl.AGARI_HAI) {
                    state.AgariHai = target;
                }
                else if (focus == FocusCtl.JIHAI) {
                    j = 13 - 3 * state.ArrayNaki.Count;
                    if (state.Jihai.Count >= j) return;
                    state.Jihai.Add(target);
                    state.Jihai.Sort(HaiCompare);
                }

            }
            Redraw();

        }

        private void MainForm_Paint(object sender, PaintEventArgs e) {
            Redraw();
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e) {
            //得点表示ウインドウを消す
            score_win.Close();
            score_win.Dispose();
        }

        private void btnClear_Click(object sender, EventArgs e) {
            //データクリア
            numChi.Value = 0;
            numPon.Value = 0;
            numMinkan.Value = 0;
            numAnkan.Value = 0;
            lstJikaze.SelectedIndex = 0;
            lstBakaze.SelectedIndex = 0;
            lstRichi.SelectedIndex = 0;
            lstTsumo.SelectedIndex = 0;
            chkChankan.Checked = false;
            chkHaitei.Checked = false;
            chkIppatsu.Checked = false;
            chkRinshan.Checked = false;

            state.AgariHai = Hai.EMPTY;
            state.ArrayNaki.Clear();
            state.Jihai.Clear();
            state.OmoteDora.Clear();
            state.UraDora.Clear();
            Redraw();
            
            
        }

    }
}